/*

# Copyright (c) 2003 Todd T. Fries <todd@fries.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 compression
*/


#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <openssl/evp.h>

#include "hzip.h"

char * show_base2(u_int8_t data);

void
hash(u_int8_t *b, size_t s, hzip_block_t *hb)
{
#if SSL
	const EVP_MD *md;
	EVP_MD_CTX mdctx;
	int md_len;
#endif


	// assume next power of 2 =< CHUNK_ZIP_SIZE
	if( (s & 0x1) == 1 ) {
		errx(1, "Oops, size is not a power of 2\n");
		// should only be here if last block
		//[ calculate next power of 2 and pad, here or calling functions? ]
   	}

#if SSL
	md = EVP_get_digestbyname("SHA1");
	if ( !md ) {
		printf("Unknown message digest SHA1\n");
		exit(1);
	}
	EVP_MD_CTX_init(&mdctx);
	EVP_DigestInit_ex(&mdctx, md, NULL);
	EVP_DigestUpdate(&mdctx, b, s);
	EVP_DigestFinal_ex(&mdctx, hb->hash, &md_len);
	EVP_MD_CTX_cleanup(&mdctx);

	printf("hash len: %d\n", md_len);
#else
	SHA1(b, s, hb->hash);
#endif

   	hb->first = b[0];
   	hb->len = revpower(s); // s = len^2
	hb->info = hzip_SHA1;

	return;
}

int
revpower(int s)
{
	int ret = 0, tmp = s;

	while( (tmp = tmp>>1) > 0 )
		ret++;

	return ret;
}

void
show_hash(u_int8_t *buf, int blocknum, size_t s)
{
	int i;
	int max = s > 30 ? 30 : s;

	if(! (debug & D_HASH))
		return;

	fprintf(stderr, "HASH[%d,%d,%d]: ",
		blocknum, s, count_bits(buf, s));

	for( i = 0; i < max; i++ ) {
		fprintf(stderr, "%02x", buf[i]);
	}

	fprintf(stderr, "\n");
}

void
show_hb(hzip_block_t *hb) {
	char *ht;
	size_t s, hs;
	double ds;
	char slen = hb->len & 0xf;
	char sinfo = hb->info & 0xf;
	char *fmt;

	if(! (debug & D_HASH))
		return;

	fprintf(stderr, "+------\n");
	switch( sinfo ) {
	case hzip_SHA1:
		ht = "SHA1";
		hs = 20;
		break;
	;;
	default:
		fprintf(stderr, "|Unknown flag: %02x\n", sinfo);
		fprintf(stderr, "+------\n");
		exit(1);
	;;
	}

	s = TWO_POWER(slen);
	fmt = "| %s len: %d size: %d first: %02x "
		"mset: %02x uset: %02x bset: %d\n| ";

	fprintf(stderr, fmt,
		ht, slen, s, hb->first, hb->mset, hb->uset, hb->bits_set);
	show_hash(hb->hash, slen, hs);
	fprintf(stderr, "+------\n");
}

u_int8_t
mask_set_add(u_int8_t mask, u_int8_t input)
{
	/* 1111 1111 - 1010 1010 = 0101 0101 */
	u_int8_t ret;
	char *m, *i, *r;
	ret = mask & input;
	if(debug & D_ADD) {
		m = show_base2(mask);
		i = show_base2(input);
		r = show_base2(ret);
		fprintf(stderr, "mask_set_add( '%s' , '%s' ) = '%s'\n",
			m, i, r);
		free(m);
		free(i);
		free(r);
	}
	return ret;
}
int
mask_set_check(u_int8_t set, u_int8_t input)
{
	/* this needs to verify at least 'set' bits are set, but other
	   bits can be set */
	char *s, *i;
	int ret;
	ret = (set & input) == set;
	if (debug & D_CHECK) {
		s = show_base2(set);
		i = show_base2(input);
		fprintf(stderr, "mask_set_check('%s', '%s') == %d\n",
			s, i, ret);
		free(s);
		free(i);
	}
	return ret;
}
int
mask_unset_check(u_int8_t set, u_int8_t input)
{
	char *s, *i;
	int ret = (set & (0xff - input)) == set;

	if (debug & D_CHECK) {
		s = show_base2(set);
		i = show_base2(input);
		/* verify that 'set' bits are not set, similar to above */
		fprintf(stderr, "mask_unset_check('%s', '%s') == %d\n",
			s, i, ret);
		free(s);
		free(i);
	}
	return  ret;
}
u_int8_t
mask_unset_add(u_int8_t mask, u_int8_t input)
{
	/* 1111 1111 - 1010 1010 = 0101 0101 */
	u_int8_t ret;
	char *m, *i, *r;
	ret = mask & ( 0xff - input );
	if(debug & D_ADD) {
		m = show_base2(mask);
		i = show_base2(input);
		r = show_base2(ret);
	
		fprintf(stderr, "mask_unset_add( '%s' , '%s' ) = '%s'\n",
			m, i, r);
		free(m);
		free(i);
		free(r);
	}
	return ret;
}

u_int16_t
count_bits(u_int8_t *buff, size_t s)
{
	int i, j, bits;
	u_int8_t mask;
	char *tmp;
	bits = 0;
	for(j = 0; j < s; j++) {
		for(i = 0; i < 8; i++) {
			mask = (0x01 << i);
			if (buff[j] & mask)
				bits++;
		}
		if (debug & D_BITS) {
			tmp = show_base2(buff[j]);
			fprintf(stderr, "count_bits: %s, %d\n", tmp, bits);
		}
	}
	return bits;
}

char *
show_base2(u_int8_t data)
{
	char buf[8];
	char *tmp;
	int i;
	u_int8_t mask;

	for(i = 0; i < 8; i++) {
		mask = (0x01 << (7-i));
		if (data & mask)
			buf[i] = '1';
		else
			buf[i] = '0';
	}
	if( (0x80 & data) || data < 0x20 ) {
		asprintf(&tmp, "%c%c%c%c %c%c%c%c",
			buf[0], buf[1], buf[2], buf[3],
			buf[4], buf[5], buf[6], buf[7]);
	} else {
		asprintf(&tmp, "%c%c%c%c %c%c%c%c(%c)",
			buf[0], buf[1], buf[2], buf[3],
			buf[4], buf[5], buf[6], buf[7],
			data);
	}
		
	return tmp;
}
