/* 

# Copyright (c) 2003 Todd T. Fries <todd@fries.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
decompression
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <openssl/evp.h>
#include <math.h>

#include "hzip.h"

u_int8_t *_hunzip(hzip_block_t *hb);
static void incr( hzip_bvars_t * );

void
hunzip(char *file)
{
	int in, out;
	hzip_block_t hb;
	size_t s;
	size_t sr;
	u_int8_t *buf;

	in = open(file, O_RDONLY, 600);
	out = open("/dev/stdout", O_RDONLY, 600);
	while( s = read(in, &hb, sizeof(hzip_block_t))) {
		if (s == -1)
			break;
		if(debug & D_BASIC) {
			fprintf(stderr, "Read %d bytes\n", s);
		}
		buf = _hunzip(&hb);
		write( out, (void *)buf, sizeof(hzip_block_t));
	}
	close(in);
	close(out);
	return;
}

u_int8_t *
_hunzip(hzip_block_t *hb) {
	hzip_block_t hb2;
	u_int64_t count = 0;
	char *ht;
	hzip_bvars_t bv;

	bv.b = hb;
	bv.s = TWO_POWER(hb->len & 0xf);

	show_hb(hb);

	bv.buff = (u_int8_t *)calloc(1, bv.s);
	bv.buff[0] = hb->first;
	bv.max = (0xff - hb->uset); /* 0xff */
	bv.bits = count_bits(bv.buff, bv.s);

	do {
		if (bv.bits == bv.b->bits_set) {
			SHA1(bv.buff, bv.s, hb2.hash);
			if ( ! memcmp(hb->hash, hb2.hash, bv.s) )
				break;
			count++;
			if (debug & D_HASH) {
				fprintf(stderr, "+====\n| ");
				show_hash(hb->hash, count, 20); // for -v
				fprintf(stderr, "| ");
				show_hash(hb2.hash, count, 20); // for -v
				fprintf(stderr, "| ");
				show_hash(bv.buff, count, bv.s);
			} else if (debug & D_BASIC)
				fprintf(stderr, ".");
		}
		incr(&bv);
		bv.bits = count_bits(bv.buff, bv.s);

			
	} while (1);
	if (debug & D_HASH) {
		fprintf(stderr, "Buffer:\n");
		show_hash(bv.buff, count, bv.s);
	}

	return bv.buff;
}

/* twiddle bits in an incrementing fashion, try to be efficient */
static void
incr( hzip_bvars_t *bv )
{
	int i;
	char *b1, *b2, *b3;
	u_int8_t us, ms;
	us = bv->b->uset;
	ms = bv->b->mset;

	/* I use '1' here because hb->first (bv.buff[0]) is already known */
	for(i = 1; i <= bv->s; i++) {
		if ( bv->buff[i] != bv->max )
			break;
		else
			bv->buff[i] = 0;
	}
	if ( i > bv->s) {
		return;
	}

	bv->bits = 0;

	/* short-circuit return if all masks are empty */

	if ( ! (ms || us || bv->b->bits_set) ) {
		bv->buff[i]++;
		return;
	}

	/* find something that works with both masks and return */
	
	while (bv->buff[i] != bv->max) {
		bv->buff[i]++;

		if (ms && us) {
			if ( mask_set_check(ms, bv->buff[i]) &&
				mask_unset_check(us, bv->buff[i]) )
				break;
			continue;
		}
		if (ms) {
			if ( mask_set_check(ms, bv->buff[i]) )
				break;
			continue;
		}
		if (us) {
			if ( mask_unset_check(us, bv->buff[i]) )
				break;
			continue;
		}

	}
	if (debug & D_INCR) {
		b1 = show_base2(bv->buff[i]);
		b2 = show_base2(bv->b->mset);
		b3 = show_base2(bv->b->uset);
		fprintf(stderr,
			"returning buff[%d] = 0x%02x ('%s'), mset = '%s', "
			"uset = '%s'\n",
			i, bv->buff[i], b1, b2, b3);
		free(b1);
		free(b2);
	}
}
