/* 

# Copyright (c) 2003 Todd T. Fries <todd@fries.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
decompression
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

#include <openssl/evp.h>
#include <openssl/sha.h>

#include "hzip.h"

typedef struct {
	u_int8_t vlist[256];
	int count;
} vc_t;

u_int8_t *_hunzip(hzip_block_t *hb, int);
static void incr( hzip_bvars_t *, vc_t * );
static vc_t * gen_table( hzip_bvars_t *bv );

void
hunzip(char *file)
{
	int in, out, c = 0;
	hzip_block_t hb;
	ssize_t s;
	u_int8_t *buf;

	in = open(file, O_RDONLY, 600);
	out = open("/dev/stdout", O_WRONLY, 600);
	while( (s = read(in, &hb, sizeof(hzip_block_t))) ) {
		if (s == -1)
			break;
		c++;
		if(debug & D_BASIC) {
			fprintf(stderr, "Read %d bytes from block %d\n", s, c);
		}
		buf = _hunzip(&hb, c);
		write( out, (void *)buf, hb.len + 1);
	}
	close(in);
	close(out);
	return;
}
	

u_int8_t *
_hunzip(hzip_block_t *hb, int c) {
	hzip_block_t hb2;
	u_int64_t count = 0;
	hzip_bvars_t bv;
	vc_t *valid_chars;
	int i;

	bv.b = hb;
	bv.s = hb->len + 1;

	show_hb(hb);

	bv.buff = (u_int8_t *)calloc(1, bv.s);
	bv.buff[0] = hb->first;
	bv.max = (0xff - hb->uset); /* 0xff */

	/* fprintf(stderr, "pre gen table\n"); */
	valid_chars = gen_table(&bv);
	/* fprintf(stderr, "post gen table\n"); */

	/* init vcp (valid char pointers) to the 1st valid char */
	/* fprintf(stderr, "pre init vcp\n"); */
	for( i = 1; i < bv.s; i++) {
		bv.vcp[i] = &valid_chars->vlist[0];
		bv.buff[i] = valid_chars->vlist[0];
	}
	/* fprintf(stderr, "post init vcp\n"); */
	bv.bits = count_bits(bv.buff, bv.s);

	do {
		if (bv.bits == bv.b->bits_set) {
			SHA1(bv.buff, bv.s, hb2.hash);
			if ( ! memcmp(hb->hash, hb2.hash, bv.s) )
				break;
			count++;
			if (debug & D_HASH) {
				fprintf(stderr, "+====\n| ");
				show_hash(hb->hash, count, 20); /* for -v */
				fprintf(stderr, "| ");
				show_hash(hb2.hash, count, 20); /* for -v */
				fprintf(stderr, "| ");
				show_hash(bv.buff, count, bv.s);
			} else if (debug & D_BASIC)
				fprintf(stderr, ".");
		}
		incr(&bv, valid_chars);
		bv.bits = count_bits(bv.buff, bv.s);
	} while (1);
	if (debug & D_HASH) {
		fprintf(stderr, "Buffer:\n");
		show_hash(bv.buff, count, bv.s);
	}
	free(valid_chars);

	fprintf(stderr, "Total hash calls for block %d: %lld\n",
			c, count);

	return bv.buff;
}

static void
incr( hzip_bvars_t *bv, vc_t *vc )
{
	int i;

	/* for each char in the buffer, iterate over vc->vlist[] */
	for(i = 1; i < bv->s; i++) {
	/*	fprintf(stderr, "incr: pos %d val 0x%x\n", i, bv->buff[i]); */
			
		if ( (bv->vcp[i] - &vc->vlist[0] + 1) > vc->count ) {
			bv->buff[i] = * (bv->vcp[i] = &vc->vlist[0]);
			continue;
		}
		bv->buff[i] = * ( (bv->vcp[i]) ++);
		break;
	}
	/* fprintf(stderr, "incr: (returning)\n"); */
}

static vc_t *
gen_table( hzip_bvars_t *bv )
{
	u_int8_t us, ms, csetmin, csetmax, bits, byte;
	vc_t *vc;
	vc = (vc_t *)calloc(1, sizeof(vc_t));
	us = bv->b->uset;
	ms = bv->b->mset;
	csetmin = bv->b->cset & 0x0f;
	csetmax = (bv->b->cset & 0xf0) >> 4;

	vc->count = 0;

	/* fprintf(stderr, "starting gen_table\n"); */

	for(byte = 0; byte < 255; byte++) {

		bits = bitcount(byte);
		if (bits > csetmax)
			continue;
		if (bits < csetmin)
			continue;

		if (ms && ! mask_set_check(ms, byte) )
			continue;
		if (us && ! mask_unset_check(us, byte) )
			continue;
		vc->vlist[vc->count++] = byte;
	}
	/* fprintf(stderr, "finished gen_table\n");	 */
	show_hash(vc->vlist, 0, vc->count);
	return vc;
}
