/*

# Copyright (c) 2003 Todd T. Fries <todd@fries.net>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

*/


#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <err.h>

#include <openssl/evp.h>
#include <openssl/sha.h>

#include "hzip.h"

char * show_base2(u_int8_t data);

void
hash(u_int8_t *b, ssize_t s, hzip_bvars_t *bv)
{
	hzip_block_t *hb = bv->b;
#if SSL
	const EVP_MD *md;
	EVP_MD_CTX mdctx;
	int md_len;
#endif


#if SSL
	md = EVP_get_digestbyname("SHA1");
	if ( !md ) {
		printf("Unknown message digest SHA1\n");
		exit(1);
	}
	EVP_MD_CTX_init(&mdctx);
	EVP_DigestInit_ex(&mdctx, md, NULL);
	EVP_DigestUpdate(&mdctx, b, s);
	EVP_DigestFinal_ex(&mdctx, hb->hash, &md_len);
	EVP_MD_CTX_cleanup(&mdctx);

	printf("hash len: %d\n", md_len);
#else
	SHA1(b, s, hb->hash);
#endif

   	hb->first = b[0];
   	hb->len = (s - 1) & 0xff; /* always < 256, len is u_int8_t */

	return;
}

int
revpower(u_int64_t s)
{
	int ret = 0;
	u_int64_t tmp = s;

	while( (tmp = tmp>>1) > 0 )
		ret++;

	return ret;
}

void
show_hash(u_int8_t *buf, u_int64_t blocknum, size_t s)
{
	int i;
	int max = s > 30 ? 30 : s;

	if(! (debug & D_HASH))
		return;
	fprintf(stderr, "HASH[%lld,%d,%d]: ",
		blocknum, s, count_bits(buf, s));

	for( i = 0; i < max; i++ ) {
		fprintf(stderr, "%02x", buf[i]);
	}

	fprintf(stderr, "\n");
}

void
show_hb(hzip_bvars_t *bv) {
	hzip_block_t *hb = bv->b;
	char *ht;
	size_t s, hs;
	u_int8_t slen = hb->len + 1;
	char *fmt;

	switch( bv->algo ) {
	case hzip_SHA1:
		ht = "SHA1";
		hs = SHA_DIGEST_LENGTH;
		break;
	;;
	default:
		fprintf(stderr, "+------\n");
		fprintf(stderr, "|Unknown flag: %02x\n", bv->algo);
		fprintf(stderr, "+------\n");
		exit(1);
	;;
	}
	if(! (debug & D_HASH))
		return;

	fprintf(stderr, "+------\n");

	s = slen;
	fmt = "| %s len: %d size: %d first: %02x "
		"mset: %02x uset: %02x bset: %d cset(%02x)\n| ";

	fprintf(stderr, fmt,
		ht, slen, s, hb->first, hb->mset, hb->uset, hb->bits_set,
		hb->cset);
	show_hash(hb->hash, slen, hs);
	fprintf(stderr, "+------\n");
}

u_int8_t
mask_set_add(u_int8_t mask, u_int8_t input)
{
	/* 1111 1111 - 1010 1010 = 0101 0101 */
	u_int8_t ret;
	char *m, *i, *r;
	ret = mask & input;
	if(debug & D_ADD) {
		m = show_base2(mask);
		i = show_base2(input);
		r = show_base2(ret);
		fprintf(stderr, "mask_set_add( '%s' , '%s' ) = '%s'\n",
			m, i, r);
		free(m);
		free(i);
		free(r);
	}
	return ret;
}
int
mask_set_check(u_int8_t set, u_int8_t input)
{
	/* this needs to verify at least 'set' bits are set, but other
	   bits can be set */
	char *s, *i;
	int ret;
	ret = (set & input) == set;
	if (debug & D_CHECK) {
		s = show_base2(set);
		i = show_base2(input);
		fprintf(stderr, "mask_set_check('%s', '%s') == %d\n",
			s, i, ret);
		free(s);
		free(i);
	}
	return ret;
}
int
mask_unset_check(u_int8_t set, u_int8_t input)
{
	char *s, *i;
	int ret = (set & (0xff - input)) == set;

	if (debug & D_CHECK) {
		s = show_base2(set);
		i = show_base2(input);
		/* verify that 'set' bits are not set, similar to above */
		fprintf(stderr, "mask_unset_check('%s', '%s') == %d\n",
			s, i, ret);
		free(s);
		free(i);
	}
	return  ret;
}
u_int8_t
mask_unset_add(u_int8_t mask, u_int8_t input)
{
	/* 1111 1111 - 1010 1010 = 0101 0101 */
	u_int8_t ret;
	char *m, *i, *r;
	ret = mask & ( 0xff - input );
	if(debug & D_ADD) {
		m = show_base2(mask);
		i = show_base2(input);
		r = show_base2(ret);
	
		fprintf(stderr, "mask_unset_add( '%s' , '%s' ) = '%s'\n",
			m, i, r);
		free(m);
		free(i);
		free(r);
	}
	return ret;
}

u_int16_t
count_bits(u_int8_t *buff, size_t s)
{
	int i, j, bits;
	u_int8_t mask;
	char *tmp;
	bits = 0;
	for(j = 0; j < s; j++) {
		for(i = 0; i < 8; i++) {
			mask = (0x01 << i);
			if (buff[j] & mask)
				bits++;
		}
		if (debug & D_BITS) {
			tmp = show_base2(buff[j]);
			fprintf(stderr, "count_bits: %s, %d\n", tmp, bits);
		}
	}
	return bits;
}

char *
show_base2(u_int8_t data)
{
	char buf[8];
	char *tmp;
	int i;
	u_int8_t mask;

	for(i = 0; i < 8; i++) {
		mask = (0x01 << (7-i));
		if (data & mask)
			buf[i] = '1';
		else
			buf[i] = '0';
	}
	if( (0x80 & data) || data < 0x20 ) {
		asprintf(&tmp, "%c%c%c%c %c%c%c%c",
			buf[0], buf[1], buf[2], buf[3],
			buf[4], buf[5], buf[6], buf[7]);
	} else {
		asprintf(&tmp, "%c%c%c%c %c%c%c%c(%c)",
			buf[0], buf[1], buf[2], buf[3],
			buf[4], buf[5], buf[6], buf[7],
			data);
	}
		
	return tmp;
}
