/* 

# Copyright (c) 2003 Todd T. Fries <todd@fries.net>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

decompression
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>

#include <openssl/evp.h>
#include <openssl/sha.h>

#include "hzip.h"

u_int8_t *_hunzip(hzip_bvars_t *, int);
static void incr( hzip_bvars_t *, vc_t *);
vc_t * gen_table( hzip_bvars_t *);
static void check_version(hzip_header_t *, hzip_bvars_t *);

void
hunzip(char *file)
{
	int in, out, c = 0;
	hzip_block_t hb;
	hzip_header_t hh;
	hzip_bvars_t bv;
	ssize_t s;
	u_int8_t *buf;

	in = open(file, O_RDONLY, 600);
	out = open("/dev/stdout", O_WRONLY, 600);

	s = read(in, &hh, sizeof(hzip_header_t));
	check_version(&hh, &bv);

	bv.b = &hb;

	while( (s = read(in, &hb, sizeof(hzip_block_t))) ) {
		if (s == -1)
			break;
		c++;
		if(debug & D_BASIC) {
			fprintf(stderr, "Read %d bytes from block %d\n", s, c);
		}
		buf = _hunzip(&bv, c);
		write( out, (void *)buf, hb.len + 1);
	}
	close(in);
	close(out);
	return;
}
	

u_int8_t *
_hunzip(hzip_bvars_t *bv, int c) {
	hzip_block_t *hb = bv->b;
	hzip_block_t hb2;
	u_int64_t count = 0;
	int i;
	u_int8_t vcount;

	bv->s = hb->len + 1;

	show_hb(bv);

	bv->buff = (u_int8_t *)calloc(1, bv->s);
	bv->buff[0] = hb->first;
	bv->max = (0xff - hb->uset); /* 0xff */

	/* fprintf(stderr, "pre gen table\n"); */
	bv->vc = gen_table(bv);
	/* fprintf(stderr, "post gen table\n"); */

	/* init vcp (valid char pointers) to the 1st valid char */
	/* fprintf(stderr, "pre init vcp\n"); */
	for( i = 1; i < bv->s; i++) {
		bv->vcp[i] = &bv->vc->vlist[0];
		bv->buff[i] = bv->vc->vlist[0];
	}
	/* fprintf(stderr, "post init vcp\n"); */
	bv->bits = count_bits(bv->buff, bv->s);

	do {
		if (bv->bits == bv->b->bits_set) {
			vcount = set_vc_count(bv);
/*
			fprintf(stderr,
				" (bits: %02x vs %02x; vcount: %02x vs %02x)\n",
				bv->bits, bv->b->bits_set, vcount,
				bv->b->vcount);
*/
			if (vcount == bv->b->vcount) {
				SHA1(bv->buff, bv->s, hb2.hash);
				count++;

				if (debug & D_HASH) {
					fprintf(stderr, "+====\n| ");
					show_hash(hb->hash, count, 20); /* for -v */
					fprintf(stderr, "| ");
					show_hash(hb2.hash, count, 20); /* for -v */
					fprintf(stderr, "| ");
					show_hash(bv->buff, count, bv->s);
				} else if (debug & D_BASIC)
					fprintf(stderr, ".");

				if ( ! memcmp(hb->hash, hb2.hash, bv->s) )
					break;
			}
		}
		incr(bv, bv->vc);
		bv->bits = count_bits(bv->buff, bv->s);
	} while (1);
	if (debug & D_HASH) {
		fprintf(stderr, "Buffer:\n");
		show_hash(bv->buff, count, bv->s);
	}
	free(bv->vc);

	fprintf(stderr, "Total hash calls for block %d: %lld\n",
			c, count);

	return bv->buff;
}

static void
incr( hzip_bvars_t *bv, vc_t *vc )
{
	int i;

	/* for each char in the buffer, iterate over vc->vlist[] */
	for(i = 1; i < bv->s; i++) {
	/*	fprintf(stderr, "incr: pos %d val 0x%x\n", i, bv->buff[i]); */
			
		if ( (bv->vcp[i] - &vc->vlist[0] + 1) > vc->count ) {
			bv->buff[i] = * (bv->vcp[i] = &vc->vlist[0]);
			continue;
		}
		bv->buff[i] = * ( (bv->vcp[i]) ++);
		break;
	}
	/* fprintf(stderr, "incr: (returning)\n"); */
}

vc_t *
gen_table( hzip_bvars_t *bv )
{
	u_int8_t us, ms, csetmin, csetmax, bits, byte;
	vc_t *vc;
	vc = (vc_t *)calloc(1, sizeof(vc_t));
	us = bv->b->uset;
	ms = bv->b->mset;
	csetmin = bv->b->cset & 0x0f;
	csetmax = (bv->b->cset & 0xf0) >> 4;

	vc->count = 0;

	/* fprintf(stderr, "starting gen_table\n"); */

	for(byte = 0; byte < 255; byte++) {

		bits = bitcount(byte);
		if (bits > csetmax)
			continue;
		if (bits < csetmin)
			continue;

		if (ms && ! mask_set_check(ms, byte) )
			continue;
		if (us && ! mask_unset_check(us, byte) )
			continue;
		vc->vlist[vc->count++] = byte;
	}
	/* fprintf(stderr, "finished gen_table\n");	 */
	show_hash(vc->vlist, 0, vc->count);
	return vc;
}

static void
check_version(hzip_header_t *hh, hzip_bvars_t *bv)
{
	if ( strncmp(hh->str, "HZIP", 5) ) {
		fprintf(stderr, "Not a hzip file!\n");
		exit(1);
	}
	bv->algo = hh->algo;
	fprintf(stderr, "Found HZIP header version %d.%d\n",
		hh->major, hh->minor);
}
