/* 

# Copyright (c) 2003 Todd T. Fries <todd@fries.net>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

compression
- sha1 based
  - pick a block size, "compress" based on
    sha1 signature
  - perhaps use some kind of info to help generate
    the data again, like number of bits, ??
  - verify that no earlier generated block of data
     generates the data, perhaps keep track of the
     match number
  - make it recognize a pattern the size of the hash,
    make hzip_PATTERN a flag(hash algorithm) for
    bits of repeating natures, implement later
  - header is 
	["hzip" + header size + filesize + pad]
      (allow for signing)
... the point is not to be fast but to
  guarantee shrinkage at the expense of cpu time
  for retrieval
    1024 bytes --> 16 bytes + overhead
*/


#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include <openssl/evp.h>

#include "hzip.h"

void
hzip(char *file)
{
	hzip_block_t hb;
	hzip_bvars_t bv;
	ssize_t s;
	int i;
	u_int8_t buff[CHUNK_ZIP_SIZE];
	int in, out;
	int blocknum = 0;
	hzip_header_t hh;

	bzero(&hh, sizeof(hzip_header_t));
	bzero(&bv, sizeof(hzip_bvars_t));
	bzero(buff, sizeof(buff));

	strlcpy(hh.str, "HZIP", sizeof(hh.str));
	hh.hlen = sizeof(hzip_header_t);  /* XXX allow this to be variable */
	bv.algo = hh.algo = hzip_SHA1;	  /* XXX allow for cmdline setting */

	hh.major = 0;
	hh.minor = 2;

	in = open(file, O_RDONLY, 600);
	if(in == -1)
		exit(1);
	out= open("/dev/stdout", O_WRONLY, 600);
	if(out == -1)
		exit(1);

	if (debug & D_BASIC) {
		fprintf(stderr, "chunk size: %d  block size: %d\n",
			CHUNK_ZIP_SIZE, sizeof(hzip_block_t));
	}

	write(out, (void *)&hh, sizeof(hzip_header_t));

	while( (s = read(in, buff, CHUNK_ZIP_SIZE)) ) {
		u_int8_t mset, uset, csetmax, csetmin;
		mset = uset = 0xff;
		csetmin = 0x8;
		csetmax = 0;

		for(i = 0; i < s; i++) {
			if (mset)
				mset = mask_set_add(mset, buff[i]);
			if (uset)
				uset = mask_unset_add(uset, buff[i]);
			if (csetmin > 0 || csetmax < 8) {
				u_int8_t bits = bitcount(buff[i]);
				if ( bits < csetmin )
					csetmin = bits;
				if ( bits > csetmax )
					csetmax = bits;
				if (debug & D_BITS)
					fprintf(stderr, "bits: %d [%d .. %d]\n",
						bits, csetmin, csetmax);
			}
		}
		hb.mset = mset;
		hb.uset = uset;
		hb.cset = (csetmax << 4) | csetmin;
		hb.bits_set = count_bits(buff, s);
		if (debug & D_BITS)
			fprintf(stderr, "cset: 0x%02x\n", hb.cset);
		bv.b = &hb;
		bv.buff = buff;
		bv.s = s;

   		hash(buff, s, &bv);
		bv.vc = gen_table(&bv);

		hb.vcount = set_vc_count(&bv);

		show_hb(&bv);

		show_hash(buff, blocknum, s);

		write(out, (void *)&hb, sizeof(hzip_block_t));
		free(bv.vc);

		blocknum++;
	}
}

u_int8_t
set_vc_count(hzip_bvars_t *bv) {
	int i, j;
	u_int8_t vcount = 0;

	for (i = 0; i < bv->vc->count; i++) {
		for (j = 0; j < bv->s; j++) {
/*
			fprintf(stderr, " (vlist[%d] = '%c'; bv->buff[%d] = '%c')\n",
			i, bv->vc->vlist[i], j, bv->buff[j]);
 */
			if (bv->vc->vlist[i] == bv->buff[j]) {
				vcount++;
				break;
			}
		}
/*
		fprintf(stderr, " (Tested [%d] = '%c', vcount = %d)\n",
			i, bv->vc->vlist[i], vcount);
 */
	}

/*
	fprintf(stderr, " (vcount = %d, bv->s = %d, bv->vc->count = %d)\n",
		 vcount, bv->s, bv->vc->count);
 */
	return vcount;
}

u_int8_t
bitcount(u_int8_t byte) {
	u_int8_t bits = 0;
	int bit;
	for(bit = 0; bit < 8; bit++)
		if (byte & (0x01 << bit))
			bits++;
	return bits;
}
