/* 

# Copyright (c) 2003 Todd T. Fries <todd@fries.net>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

*/

#include <ssl/sha.h>

#define CHUNK_ZIP_SIZE	256

struct hzip_header; /* struct written to the file once */
struct hzip_block;  /* struct written to the file many times */
struct hzip_bvars;  /* internal usage struct for passing vars and hzip_blocks */

struct valid_chars;

typedef struct hzip_header hzip_header_t;
typedef struct hzip_block hzip_block_t;
typedef struct hzip_bvars hzip_bvars_t;
typedef struct valid_chars vc_t;

struct hzip_header {
	char str[5];
	u_int8_t hlen;		/* = sizeof(hzip_header_t); */
	u_int8_t slen;		/* stream length */
	u_int8_t algo;		/* hash algorithm */
	u_int8_t major;		/* major version number */
	u_int8_t minor;		/* minor version number */
	u_int8_t data[13]; /* obviously this can be increased in size */
};

/*  Once things mature, instead of always starting at the beginning, reduce
    the number of computations by 4 again by specifying direction and beginning
    point
#define hzip_TOPDOWN	0x10
#define hzip_MIDUP	0x20
#define hzip_MIDDOWN	0x30
#define hzip_BOTTOMUP	0x40
 */

/* thought: force hash type on all blocks so hash type is in the header not
   the block struct?? */

/* for algo fields */
#define hzip_SHA1	0x01
#define hzip_SHA256	0x02
#define hzip_SHA512	0x03
#define hzip_MD5	0x04

struct hzip_block {
	u_int8_t hash[/* EVP_MAX_MD_SIZE */ SHA_DIGEST_LENGTH];
	u_int8_t first;

	u_int8_t len; /* len + 1 = block size (never zero length buffer) */

	u_int8_t iter; 	/* number of hash collisions - 1 ( 0 = 1st found,  */
			/*				   2 = 2nd found)  */
	u_int8_t mset;	/* set bits bitmask to speed computation */
	u_int8_t uset; /* unset bits bitmask to speed computation */
	u_int16_t bits_set; /* count number of bits set to speed computation */
	u_int8_t cset;  /* 1st 4 bits: max # of bits set per char */
			/* 2nd 4 bits: min # of bits set per char */
	u_int8_t vcount; /* valid char count */
};

/*
11111111 = 8 bits set

0000 = 0
0001 = 1
0010 = 2
0011 = 3
0100 = 4
0101 = 5
0110 = 6
0111 = 7
1000 = 8
*/

struct hzip_bvars {
	u_int8_t *buff;
	u_int8_t *vcp[CHUNK_ZIP_SIZE];
	size_t s;
	hzip_block_t *b;
	u_int8_t max;
	u_int16_t bits;
	u_int8_t algo;		/* hash zip algorithm */
	vc_t *vc;
};

struct valid_chars {
	u_int8_t vlist[256];
	int count;
};



#define TWO_POWER(x)	(1<<x)

/* #define CHUNK_ZIP_POWER	10 */
/* #define CHUNK_ZIP_POWER	7 */
/* #define CHUNK_ZIP_SIZE TWO_POWER(CHUNK_ZIP_POWER) */
/* always a Power of 2 */

extern int debug;

#define D_BASIC	(1<<0)
#define D_HASH	(1<<1)
#define D_ADD	(1<<2)
#define D_CHECK	(1<<3)
#define D_BITS	(1<<4)
#define D_INCR	(1<<5)

void hash(u_int8_t *, ssize_t, hzip_bvars_t *);
void hzip(char *file);
void hunzip(char *file);
void show_hash(u_int8_t *buf, u_int64_t blocknum, size_t s);
void show_hb(hzip_bvars_t *);
char *show_base2(u_int8_t);

int revpower(u_int64_t);

u_int8_t mask_set_add(u_int8_t, u_int8_t);
int mask_set_check(u_int8_t, u_int8_t);
u_int8_t mask_unset_add(u_int8_t, u_int8_t);
int mask_unset_check(u_int8_t, u_int8_t);
u_int16_t count_bits(u_int8_t *, size_t);
u_int8_t bitcount(u_int8_t);

vc_t * gen_table( hzip_bvars_t *);


u_int8_t set_vc_count(hzip_bvars_t *);
